# -*- coding: utf-8 -*-
"""
ChE 210D F19 
Final Project
12/11/2019
@author: Tobias Mazal
"""

#import python modules
import numpy as np
from scipy.stats import linregress
import matplotlib.pyplot as plt  

#import sys
#import os
#import time
import atomwrite

#import compiled fortran library
import projectlib
#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
#set the random number seed; useful for debugging
#np.random.seed(342324)

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

AtomType = 1

def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    #start the iteration counter
    Iter = 0
    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [projectlib.calcenergy(Pos + NormDir * x, M, L, Cut, NSolvent, NWall, zj, kbias) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = projectlib.calcenergy(Pos + NormDir * dx, M, L, Cut, NSolvent, NWall, zj, kbias)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [projectlib.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut, NSolvent, NWall, zj, kbias)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [projectlib.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut, NSolvent, NWall, zj, kbias)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print("Warning: maximum number of iterations reached in line search.")
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = projectlib.calcenergy(Pos + NormDir * d3, M, L, Cut, NSolvent, NWall, zj, kbias)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = projectlib.calcenergy(Pos + NormDir * d3, M, L, Cut, NSolvent, NWall, zj, kbias)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = projectlib.calcenergyforces(Pos, M, L, Cut, NSolvent, NWall, zj, kbias, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = projectlib.calcenergyforces(Pos, M, L, Cut, NSolvent, NWall, zj, kbias, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience. Not used in this code.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    Pos[:,2] = int(0.5*L)
                    return Pos
    Pos[:,2] = int(0.5*L)
    return Pos


def InitWallAndSolventPos(L,Width):
    """Returns an array of initial positions of each wall and surface atom.
       Wall atoms are placed on FCC lattice for convenience. 
Input:
    L: box length
    Width: Width of crystal lattice
Output:
    PosSolvent: (NLat-Width,3) array of positions of atoms on the wall
    PosWall: (Width,3) array of positions of atoms on the wall
    
"""
    #Set length and width of wall
    NLat = L                #Length = box length
    Spacing = 2             #Spacing of solvent molecules. Inversely proportional to concentration
    LatSpac = (L / NLat)
    SolventFlag = True      #Place solvent?
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    r2 = LatSpac * np.arange(2*NLat, dtype=float)/2 - 0.5*L
    PosWall = np.zeros((2*Width*NLat**2,3),float)
    if SolventFlag:
        PosSolvent = np.zeros((((NLat-Width)*NLat**2)/Spacing,3),float)
    else:
        PosSolvent = np.zeros((0,3),float)
    #loop through x, y, z positions in lattice until done
    #for every atom on the wall
    i = 0
    j = 0
    for x in r:
        for y in r:
            #Place wall atoms
            for z in range(Width):
                PosWall[i] = np.array([x,y,r[z]], float) 
                i += 1
                PosWall[i] = np.array([x+0.5,y+0.5,r2[2*z+1]], float) #offset for BCC                
                i += 1
            #Place solvent atoms
            if SolventFlag:
                for z in range(NLat - Width):
                    if j%Spacing == 0 and j < (NLat-Width)*NLat**2-1: #Place solvent molecules at a given spacing
                        PosSolvent[j/Spacing] = np.array([x,y,r[z+Width]])
                    j = j+1
    return PosSolvent,PosWall

def RescaleVelocities(Vel, NWall, T):
    """Rescales velocities in the system to the target temperature.
Input:
    NWall: number of wall molecules
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    if NWall != 0: #properly set scale if no wall exists
        VScale = np.sqrt(1.5 * len(Vel[:-NWall]) * T / KE)
    else:
        VScale = np.sqrt(1.5 * len(Vel) * T / KE)    
    Vel = Vel * VScale
    if NWall != 0: #Zero wall velocities
        Vel[-NWall:] = 0
    return Vel  

def InitVelocities(N, NWall, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    NWall: numer of wall atoms (not moving)
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, NWall, T)
    if NWall != 0:
        Vel[-NWall:] = 0
    return Vel

def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut#, NSolvent, NWall, zj, k
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    #print(M, L, Cut, NSolvent, NWall, zj, kbias)
    PEnergy, Accel = projectlib.calcenergyforces(Pos, M, L, Cut, NSolvent, NWall, zj, kbias, Forces)
    return Accel

def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

def InitSolutePos(L,NSolute):
    """Returns an array of initial positions of each atom.
Input:
    N: number of solute atoms
    L: box length
Output:
    PosSolute: (N,3) array of solute positions
"""
    PosSolute = np.zeros((NSolute,3),float)
    PosSolute = np.random.uniform(low=-L/2, high=L/2, size=(NSolute,3))
    #PosSolute[:,2] = int(L*0.25)
    PosSolute[:,2]=0.75 #Furthest possible distance from the crystal surface, given it has a width of 1.5
    return PosSolute

L = 11 #Set min image box length (11 is target for N ~ 1000)

#Place single solute atom at initial position
NSolute = 32
NamesSolute = ["W" for x in range(NSolute)]
PosSolute = InitSolutePos(L,NSolute)

Writing = True      #Want to write to pdb?
WallFlag = True     #Want to add wall?
Width = 2           #Wall width | 2 layers
if not WallFlag:    
    Width = 0
PosSolvent, PosWall = InitWallAndSolventPos(L, Width)   #Initialize wall and solvent atom positions
#Store initial solvent atom positions
NSolvent = PosSolvent[:,0].size
NamesSolvent = ["S" for x in range(NSolvent)]
#Store wall atom positions
NWall = PosWall[:,0].size
NamesWall = ["W" for x in range(NWall)]

N = NSolute+NSolvent+NWall
Names = NamesSolute+NamesSolvent+NamesWall
Pos = np.concatenate((PosSolute,PosSolvent,PosWall), axis = 0) 
if Writing:     #Write to pdb
    Pdb = atomwrite.pdbfile("Proj.pdb",L,AtomNames=Names)
    Pdb.write(Pos)

Cut = 2.5
zj = 0
dx = 0.001
dt = 0.001
EFracTolLS = 1.e-8
EFracTolCG = 1.e-10
kbias = 12.         #biasing spring constant

NStepsEquil1 = 20000
NStepsProd = 100000
RescaleFreq = 100
RecordFreq = 500
Run = 0                     #Track which run we're on
MVals = [1,2,4]             #Vary M at T = ?
M = MVals[1]
TVals = [1,2,3]                 #Vary T at M = ?
#TVals = [0.5, 1, 1.5]

PMFFlag = False      #Are we computing PMF?
#zjVals = np.arange(-L + 4, L - 3.9, dtype=float)/2
#zjVals = np.linspace(-L/2. + 2, L/2. - 0.5, 25)
#TVals = zjVals #temp assignment for indexing purposes

#save some time by saving energy minimized state off the bat
PEnergy, Pos = ConjugateGradient(Pos,dx,EFracTolLS,EFracTolCG) #Minimize energy using conjugate gradient search
PosStored = Pos.copy()

#initialize arrays
PEnergies = np.array([[0 for x in range(NStepsEquil1+NStepsProd)] for y in range(len(TVals))], float)
KEnergies = np.array([[0 for x in range(NStepsEquil1+NStepsProd)] for y in range(len(TVals))], float)
Energies = np.array([[0 for x in range(NStepsEquil1+NStepsProd)] for y in range(len(TVals))], float)
AvgE = np.array([0 for y in range(len(TVals))], float)
TargetKE = np.array([0 for y in range(len(TVals))], float)
RunPEs = np.array([[0 for x in range(NStepsProd/RecordFreq)] for y in range(len(TVals))], float)
RunZs = np.array([[0 for x in range(NStepsProd/RecordFreq)*NSolute] for y in range(len(TVals))], float) #Counter for z values!
MSDs = np.array([[0 for x in range(NStepsProd/RecordFreq)] for y in range(len(TVals))], float)
DiffConstants = np.array([0 for y in range(len(TVals))], float)

ZMin = 0         #technical min
ZMax = (L-1.5)/2. #technical max
Bins = 25
BinWidth = (ZMax - ZMin)/Bins
if PMFFlag:
    ProbZj = np.array([[0 for x in range(Bins)] for y in range(len(zjVals))], float)
    EtaZj = np.array([[0 for x in range(Bins)] for y in range(len(zjVals))], float)
    BetaFz = np.array([[0 for x in range(Bins)] for y in range(len(zjVals))], float)
    BinSpan = np.array([[0 for x in range(Bins+1)] for y in range(len(zjVals))], float)

for x in TVals:
    Pos = PosStored #refresh initial positions
    i = 0       #Counter for equilibration
    k = 0       #Counter for production run
    T = x       #set T to run temp
    Vel = InitVelocities(N,NWall,T)       #Take initial velocities and accels
    Accel = InitAccel(Pos)
    if Writing:
        Pdb.write(Pos)
    while i < NStepsEquil1:             #Integrate using velocity verlet first time
        Pos, Vel, Accel, KEnergy, PEnergy = projectlib.vvintegrate(Pos,Vel,Accel,M,L,Cut,dt,NSolvent,NWall,zj,kbias)
        PEnergies[Run][i] = PEnergy
        KEnergies[Run][i] = KEnergy
        i = i+1
        if (i % RescaleFreq == 0):              #Is it time to rescale vels?
            Vel = RescaleVelocities(Vel, NWall, T)
            if (i % RecordFreq == 0 and Writing):
                Pdb.write(Pos)
    Energies[Run][:NStepsEquil1] = PEnergies[Run][:NStepsEquil1] + KEnergies[Run][:NStepsEquil1]
    AvgE[Run] = np.mean(Energies[Run][-(NStepsProd+NStepsEquil1):-(NStepsProd)])  #CHECK THIS!
    TargetKE[Run] = AvgE[Run] - PEnergies[Run][NStepsEquil1-1] #Find target KE
    Vel = Vel*(TargetKE[Run]/KEnergies[Run][NStepsEquil1-1])**(0.5) #Scale velocities according to target KE
    #AchievedKE = 0.5*np.sum(Vel*Vel) #For troubleshooting, check
    Pos0 = Pos[0:NSolute,:].copy() #Copy atomic positions for squared displacement calculation
    while k < NStepsProd:             #Integrate using velocity verlet second time?
        Pos, Vel, Accel, KEnergy, PEnergy = projectlib.vvintegrate(Pos,Vel,Accel,M,L,Cut,dt,NSolvent,NWall,zj,kbias)
        #Store energies
        PEnergies[Run][i+k] = PEnergy
        KEnergies[Run][i+k]= KEnergy
        Energies[Run][i+k]= KEnergy+PEnergy
         #Calculate SD for each atom and average these for MSD
        if (k % RecordFreq == 0):
            Diffs = Pos[0:NSolute,:] - Pos0
            SDs = np.square(Diffs).sum(axis=1)
            MSDs[Run][k/RecordFreq] = np.mean(SDs)   #Calculate MSDs
            if Writing:
                Pdb.write(Pos)
            print("%.2f Percent Complete" % (float(k)/NStepsProd)) #For my information :)
            for l in range(NSolute):
                if Width != 0: 
                    if Pos[l,2] <= (Width-0.5)/2: #Place z values as distance from surface
                        RunZs[Run][k/RecordFreq*NSolute+l] = abs(-Pos[l,2] - L/2. + (Width-0.5))
                    else:
                        RunZs[Run][k/RecordFreq*NSolute+l] = L/2. - Pos[l,2]
                else: #no surface, really just for my own analysis
                    RunZs[Run][k/RecordFreq*NSolute+l] = min(L/2. - Pos[l,2], Pos[l,2] + L/2.)
                
        k = k+1
    #Bulk Diff. Calc
    DiffConstants[Run] = (linregress(np.arange(0,NStepsProd/RecordFreq)/2.,MSDs[Run])[0])/6. #Calculate D as slope/6; MSDs = 6*D*t
    #Surface Diff. Calc
    #DiffConstants[Run] = (linregress(np.arange(0,NStepsProd/RecordFreq)/2.,MSDs[Run])[0])/4.
    if PMFFlag and not np.isnan(Pos[0][0]): #if recording PMF
        plt.figure(2)
        ProbZj[Run] = plt.hist(RunZs[Run],bins = Bins)[0]/RunZs[Run].size #Calculated weighted normalized probability distribution
        EtaZj[Run] = kbias/2.*(np.mean(RunZs[Run])-zj)**2 #Calculate bias for given zj
        BinSpan[Run] = plt.hist(RunZs[Run],bins = Bins)[1] #Store the range of zs covered
    Run += 1
    
if Writing:     #Close pdb file
    Pdb.close()

#Plot MSD vs. t
plt.figure(0)
plt.xlabel('MD Steps (1000)')
plt.ylabel('MSD')
for x in range(len(TVals)):
    plt.plot(np.arange(0,NStepsProd/RecordFreq)/2.,MSDs[x], label = ("M = %i" %M, "T = %i" %TVals[x]))    
plt.legend(loc = 'upper left')

#Plot PE vs. t
plt.figure(1)
plt.xlabel('MD Steps')
plt.ylabel('PE')
for x in range(len(TVals)):
    plt.plot(np.arange(NStepsEquil1+NStepsProd),PEnergies[x], label = ("M = %i" %M, "T = %i" %TVals[x]))    
plt.legend(loc = 'upper left')

if PMFFLag:
    plt.figure(3)
    plt.xlabel('z')
    plt.ylabel('beta*F(z)')
    for x in range(len(zjVals)): 
        if not np.isnan(ProbZj[x][0]):
            BetaFz[x] = -np.log(ProbZj[x]) + EtaZj[x] #calculate PMF!
            #BetaFz[x] = -np.log(ProbZj[np.nonzero(ProbZj)]) + EtaZj[np.nonzero(EtaZj)]
            plt.plot(np.transpose((BinSpan[x,1:]+BinSpan[x,0:25])/2),BetaFz[x]) #plot PMF vs z.
#plt.figure(2)
#AllRunZs = np.concatenate(RunZs)
##plt.hist(np.transpose(RunZs),bins=Bins)
#plt.hist(AllRunZs,bins=50)
#
#ProbZ = plt.hist(AllRunZs,bins=Bins)[0]/AllRunZs.size #give histogram counts, normalized for probability distribution
#etaj = 
#betaFZ = -np.log(ProbZ) + 
#

#WHAM stuff
#U_k = UMin + (k + 1/2)*BinWidth, k - bin index
#C_k = counts of U observations in the range [U_k - BinWidth/2, U_k + BinWidth/2]
#C(U) = C_k; k = int((U - UMin)/BinWidth)
